/* 
 * main.c --
 *
 *	This file contains the main program for "wish", a windowing
 *	shell based on Tk and Tcl.  It also provides a template that
 *	can be used as the basis for main programs for other Tk
 *	applications.
 *
 * Copyright (c) 1990-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#include "prototypes.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>


#define MAXCLINE 156

/*
 * Declarations for various library procedures and variables (don't want
 * to include tkInt.h or tkConfig.h here, because people might copy this
 * file out of the Tk source directory to make their own modified versions).
 */

extern void		exit _ANSI_ARGS_((int status));
extern int		isatty _ANSI_ARGS_((int fd));
/*extern int		read _ANSI_ARGS_((int fd, char *buf, size_t size));*/
extern char *		strrchr _ANSI_ARGS_((CONST char *string, int c));
extern char TCL_LIBS[];
extern char WWW_UI[];
extern char NT_UI[];
extern char GENERIC[];
/*
 * Global variables used by the main program:
 */

static Tk_Window mainWindow;	/* The main window for the application.  If
				 * NULL then the application no longer
				 * exists. */
Tcl_Interp *interp;	/* Interpreter for this application. */
#ifdef NOTDEF
/*hangover from Tk3.6*/
char *tcl_RcFileName = NULL;	/* Name of a user-specific startup script
				 * to source if the application is being run
				 * interactively (e.g. "~/.wishrc").  Set
				 * by Tcl_AppInit.  NULL means don't source
				 * anything ever. */
#endif
static Tcl_DString command;	/* Used to assemble lines of terminal input
				 * into Tcl commands. */
static int tty;			/* Non-zero means standard input is a
				 * terminal-like device.  Zero means it's
				 * a file. */
static char errorExitCmd[] = "exit 1";

/*
 * Command-line options:
 */

static int synchronize = 0;
static char *fileName = NULL;
static char *name = NULL;
static char *display = NULL;
static char *geometry = NULL;

static Tk_ArgvInfo argTable[] = {
    {"-file", TK_ARGV_STRING, (char *) NULL, (char *) &fileName,
	"File from which to read commands"},
    {"-geometry", TK_ARGV_STRING, (char *) NULL, (char *) &geometry,
	"Initial geometry for window"},
    {"-display", TK_ARGV_STRING, (char *) NULL, (char *) &display,
	"Display to use"},
    {"-name", TK_ARGV_STRING, (char *) NULL, (char *) &name,
	"Name to use for application"},
    {"-sync", TK_ARGV_CONSTANT, (char *) 1, (char *) &synchronize,
	"Use synchronous mode for display server"},
    {(char *) NULL, TK_ARGV_END, (char *) NULL, (char *) NULL,
	(char *) NULL}
};

/*
 * Declaration for Tcl command procedure to create demo widget.  This
 * procedure is only invoked if SQUARE_DEMO is defined.
 */

extern int SquareCmd _ANSI_ARGS_((ClientData clientData,
	Tcl_Interp *interp, int argc, char *argv[]));


int
ui_init(argc, argv, wname)
    int argc;				/* Number of arguments. */
    char **argv;			/* Array of argument strings. */
    char *wname;
{
    char *p;
    char buf[MAXCLINE];
    int i;

    putenv("TCL_LIBRARY=");
    putenv("TK_LIBRARY=");

    interp = Tcl_CreateInterp();
#ifdef TCL_MEM_DEBUG
    Tcl_InitMemory(interp);
#endif

    /*
     * Parse command-line arguments.
     */

    if (Tk_ParseArgv(interp, (Tk_Window) NULL, &argc, argv, argTable, 0)
	    != TCL_OK) {
	fprintf(stderr, "%s\n", interp->result);
	exit(1);
    }
    if (name == NULL) {
	if (fileName != NULL) {
	    p = fileName;
	} else {
	    p = argv[0];
	}
	name = strrchr(p, '/');
	if (name != NULL) {
	    name++;
	} else {
	    name = p;
	}
    }

    /*
     * If a display was specified, put it into the DISPLAY
     * environment variable so that it will be available for
     * any sub-processes created by us.
     */

    if (display != NULL) {
	Tcl_SetVar2(interp, "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    }
    Tcl_SetVar(interp, "auto_path", ".", TCL_GLOBAL_ONLY);
    sprintf(buf, "-name nte");
    for(i=1;i<argc;i++)
      {
	strncat(buf, " ", MAXCLINE-strlen(buf));
	strncat(buf, argv[i], MAXCLINE-strlen(buf));
      }
    Tcl_SetVar(interp, "argv", buf, TCL_GLOBAL_ONLY);
#ifndef WIN32
    if(Tk_Init(interp)!=TCL_OK)
      {
	/*I'd like to check this, but can't do anything sensible because
	  Tk_Init trys to read the TK_LIBRARY files, and I don't want it to!*/
	fprintf(stderr, "nt: error opening display: %s\n", interp->result);
	exit(1);
      }
#else
	Tcl_Init(interp);
	Tk_Init(interp);
#endif

    sprintf(buf, "wm title . \"%s\"", wname);
    Tcl_Eval(interp, buf);

    /*
     * Make command-line arguments available in the Tcl variables "argc"
     * and "argv".  Also set the "geometry" variable from the geometry
     * specified on the command line.
     */

    sprintf(buf, "%d", argc-1);
    Tcl_SetVar(interp, "argc", buf, TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "argv0", (fileName != NULL) ? fileName : argv[0],
	    TCL_GLOBAL_ONLY);
    if (geometry != NULL) {
	Tcl_SetVar(interp, "geometry", geometry, TCL_GLOBAL_ONLY);
    }

    /*
     * Set the "tcl_interactive" variable.
     */

    tty = isatty(0);
    Tcl_SetVar(interp, "tcl_interactive",
	    ((fileName == NULL) && tty) ? "1" : "0", TCL_GLOBAL_ONLY);

    /*
     * Add a few application-specific commands to the application's
     * interpreter.
     */

#ifdef SQUARE_DEMO
    Tcl_CreateCommand(interp, "square", SquareCmd, (ClientData) mainWindow,
	    (void (*)()) NULL);
#endif
    Tcl_CreateCommand(interp, "ui_quit", ui_quit, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_new_block", ui_new_block, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_delete_block", ui_delete_block, 
		      (ClientData) mainWindow, (void (*)()) NULL);
	/* extra jim command*/
	Tcl_CreateCommand(interp, "ui_undelete_block", ui_undelete_block, 
		      (ClientData) mainWindow, (void (*)()) NULL);
	/* end jim command*/
    Tcl_CreateCommand(interp, "ui_new_line", ui_new_line,
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_line", ui_set_line, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_delete_line", ui_delete_line, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_find_line_len", ui_find_line_len, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_find_block_len", ui_find_block_len, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_findblock", ui_findblock, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_load_text_file", ui_load_text_file, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_load_struct_file", ui_load_struct_file, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_save_text_file", ui_save_text_file, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_save_struct_file", ui_save_struct_file, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_generate_id", ui_generate_id, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_my_colour", ui_set_my_colour, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_block_colour", ui_set_block_colour, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_block_font", ui_set_block_font, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_block_posn", ui_set_block_posn, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_print_ds", print_ds, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_get_creator", ui_get_creator, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_get_modifier", ui_get_modifier, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_get_modification_time", 
		      ui_get_modification_time, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_get_lock_status", 
		      ui_get_lock_status, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_get_my_lock_status", 
		      ui_get_my_lock_status, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_set_lock_mode", 
		      ui_set_lock_mode, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_send_shared_pointer", 
		      ui_send_shared_pointer, 
		      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "ui_keep_userlist", 
		      ui_keep_userlist, 
		      (ClientData) mainWindow, (void (*)()) NULL);

    /*register WWW stuff*/
    Tcl_CreateCommand(interp, "webto",
                      webto,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "save_www_data_to_file",
                      save_www_data_to_file,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "stop_www_loading",
                      stop_www_loading,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "getustimeofday",
                      uigetustimeofday,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "get_participant_by_number",
                      ui_get_participant_by_number,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "set_encryption_key",
                      ui_set_encryption_key,
                      (ClientData) mainWindow, (void (*)()) NULL);
    Tcl_CreateCommand(interp, "reset_participants",
                      ui_reset_participants,
                      (ClientData) mainWindow, (void (*)()) NULL);

    /*
     * Invoke application-specific initialization.
     */


    /*
     * Loop infinitely, waiting for commands to execute.  When there
     * are no windows left, Tk_MainLoop returns and we exit.
     */

/*    Tk_MainLoop();*/
#ifdef NOTDEF
    while (tk_NumMainWindows > 0) 
      {
        Tk_DoOneEvent(0);
      }

    /*
     * Don't exit directly, but rather invoke the Tcl "exit" command.
     * This gives the application the opportunity to redefine "exit"
     * to do additional cleanup.
     */

    Tcl_Eval(interp, "exit");
    exit(1);
#endif
    return 1;

}

int ui_create_interface(int rx_fd, int participants)
{
    int code;
    char *msg;
    char part_str[80];

#ifndef WIN32
    if (Tcl_AppInit(interp) != TCL_OK) {
	  fprintf(stderr, "Tcl_AppInit failed: %s\n", interp->result);
    }
#endif
    if (Tcl_VarEval(interp, TCL_LIBS, NULL) != TCL_OK) {
	  fprintf(stderr, "TCL_LIBS failed: %s\n", interp->result);
    };

    /*
     * Set the geometry of the main window, if requested.
     */

    if (geometry != NULL) {
	code = Tcl_VarEval(interp, "wm geometry . ", geometry, (char *) NULL);
	if (code != TCL_OK) {
	    fprintf(stderr, "%s\n", interp->result);
	}
    }

    /*
     * Invoke the script specified on the command line, if any.
     */

    if (fileName != NULL) {
	code = Tcl_VarEval(interp, "source ", fileName, (char *) NULL);
	if (code != TCL_OK) {
	    goto error;
	}
	tty = 0;
    } else {
	/*
	 * Commands will come from standard input, so set up an event
	 * handler for standard input.  If the input device is aEvaluate the
	 * .rc file, if one has been specified, set up an event handler
	 * for standard input, and print a prompt if the input
	 * device is a terminal.
	 */

#ifdef NOTDEF
	if (tcl_RcFileName != NULL) {
	    Tcl_DString buffer;
	    char *fullName;
	    FILE *f;
    
	    fullName = Tcl_TildeSubst(interp, tcl_RcFileName, &buffer);
	    if (fullName == NULL) {
		fprintf(stderr, "%s\n", interp->result);
	    } else {
		f = fopen(fullName, "r");
		if (f != NULL) {
		    code = Tcl_EvalFile(interp, fullName);
		    if (code != TCL_OK) {
			fprintf(stderr, "%s\n", interp->result);
		    }
		    fclose(f);
		}
	    }
	    Tcl_DStringFree(&buffer);
	}
#endif 
	if(participants!=-1)
	  {
	    sprintf(part_str, "option add *participants %d interactive", 
		    participants);
	    Tcl_VarEval(interp, part_str, NULL);
	  }

	if(Tcl_VarEval(interp, GENERIC, NULL)!=TCL_OK)
	  {
	    debug("Tcl_VarEval GENERIC: %s\n", interp->result);
	  }
        ;
        if(Tcl_VarEval(interp, WWW_UI, NULL)!=TCL_OK)
	  {
	    Tcl_AddErrorInfo(interp, "\n");
	    debug( "WWW_UI: %s\n", interp->result);
	    Tcl_VarEval(interp, "puts $errorInfo", NULL);
	  }
	if(Tcl_VarEval(interp, NT_UI, NULL)!=TCL_OK)
	  {
	    Tcl_AddErrorInfo(interp, "\n");
	    debug("NT_UI: %s\n", interp->result);
	    Tcl_VarEval(interp, "puts $errorInfo", NULL);
	  }
#ifdef NOTDEF
	Tk_CreateFileHandler(0, TK_READABLE, StdinProc, (ClientData) 0);
	if (tty) {
	    Prompt(interp, 0);
	}
#endif
    }
    fflush(stdout);
    Tcl_DStringInit(&command);

    /*
     * Add a filehandler for the network receive socket
     */
#ifdef NOTDEF
    Tcl_CreateFileHandler(Tcl_GetFile((ClientData)rx_fd, TCL_UNIX_FD), TK_READABLE, (Tk_FileProc *)xsockrecv, (ClientData) 0);
#endif
    linksocket(rx_fd, TK_READABLE, (Tk_FileProc *)xsockrecv);  /* jim fix to cast pointer*/
    return 0;

error:
    msg = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    if (msg == NULL) {
	msg = interp->result;
    }
    fprintf(stderr, "%s\n", msg);
    Tcl_Eval(interp, errorExitCmd);
    return 1;			/* Needed only to prevent compiler warnings. */
}
